# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import shutil
import sys
from collections import OrderedDict
from typing import Dict, List, Optional, Union

import click
from texttable import Texttable

from silkaj import constants

VERT_TABLE_CHARS = ["─", "│", "│", "═"]


def send_doc_confirmation(document_name: str) -> None:
    if not click.confirm(f"Do you confirm sending this {document_name}?"):
        sys.exit(constants.SUCCESS_EXIT_STATUS)


class Table(Texttable):
    def __init__(
        self,
        style="default",
    ):
        super().__init__(max_width=shutil.get_terminal_size().columns)

        if style == "columns":
            self.set_deco(self.HEADER | self.VLINES | self.BORDER)
        self.set_chars(VERT_TABLE_CHARS)

    def fill_rows(self, rows: List[List], header: Optional[List] = None) -> None:
        """
        Fills a table from header and rows list.
        `rows` is a list of lists representing each row content.
        each element of `rows` and header must be of same length.
        """
        if header:
            if len(rows) == 0:
                rows.append([""] * len(header))
            assert len(header) == len(rows[0])
            self.header(header)
        for line in rows:
            assert len(line) == len(rows[0])
            self.add_row(line)

    def fill_from_dict(self, _dict: Union[Dict, OrderedDict]) -> None:
        """
        Given a dict where each value represents a column,
        fill a table where labels are dict keys and columns are dict values
        This function stops on the first line with only empty cells
        """
        keys = list(_dict.keys())
        rows = []

        n = 0
        while True:
            row = []
            empty_cells_number = 0

            for key in keys:
                try:
                    row.append(_dict[key][n])
                except IndexError:
                    row.append("")
                    empty_cells_number += 1
                # break on first empty row
            if empty_cells_number == len(keys):
                break
            rows.append(row)
            n += 1

        return self.fill_rows(rows, keys)

    def fill_from_dict_list(self, dict_list: List[Dict]) -> None:
        """
        Given a list of dict with same keys,
        fills the table with keys as header
        """
        header = list(dict_list[0].keys())
        content = []
        for _dict in dict_list:
            assert list(_dict.keys()) == header
            line = []
            for head in header:
                line.append(_dict[head])
            content.append(line)
        return self.fill_rows(content, header)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import time
from collections import OrderedDict
from operator import itemgetter
from urllib.error import HTTPError

from click import IntRange, argument, command, option
from duniterpy.api import bma
from pendulum import from_timestamp

from silkaj import tui
from silkaj.blockchain.tools import get_head_block
from silkaj.constants import ALL, BMA_SLEEP
from silkaj.network import client_instance
from silkaj.wot.tools import identity_of


@command("blocks", help="Display blocks: default: 0 for current window size")
@argument("number", default=0, type=IntRange(0, 5000))
@option(
    "--detailed",
    "-d",
    is_flag=True,
    help="Force detailed view. Compact view happen over 30 blocks",
)
def list_blocks(number: int, detailed: bool) -> None:
    head_block = get_head_block()
    current_nbr = head_block["number"]
    if number == 0:
        number = head_block["issuersFrame"]
    client = client_instance()
    blocks = client(bma.blockchain.blocks, number, current_nbr - number + 1)
    issuers = []
    issuers_dict = {}
    for block in blocks:
        issuer = OrderedDict()
        issuer["pubkey"] = block["issuer"]
        if detailed or number <= 30:
            gentime = from_timestamp(block["time"], tz="local").format(ALL)
            mediantime = from_timestamp(block["medianTime"], tz="local").format(ALL)
            issuer["block"] = block["number"]
            issuer["gentime"] = gentime
            issuer["mediantime"] = mediantime
            issuer["hash"] = block["hash"][:10]
            issuer["powMin"] = block["powMin"]
        issuers_dict[issuer["pubkey"]] = issuer
        issuers.append(issuer)
    for pubkey in issuers_dict.keys():
        issuer = issuers_dict[pubkey]
        time.sleep(BMA_SLEEP)
        try:
            idty = identity_of(issuer["pubkey"])
        except HTTPError:
            idty = None
        for issuer2 in issuers:
            if (
                issuer2.get("pubkey") is not None
                and issuer.get("pubkey") is not None
                and issuer2["pubkey"] == issuer["pubkey"]
            ):
                issuer2["uid"] = idty["uid"] if idty else None
                issuer2.pop("pubkey")
    print_blocks_views(issuers, current_nbr, number, detailed)


def print_blocks_views(issuers, current_nbr, number, detailed):
    header = (
        f"Last {number} blocks from n°{current_nbr - number + 1} to n°{current_nbr}"
    )
    print(header, end=" ")
    if detailed or number <= 30:
        sorted_list = sorted(issuers, key=itemgetter("block"), reverse=True)

        table = tui.Table(style="columns")
        table.set_cols_align(["r", "r", "r", "r", "r", "l"])
        table.set_cols_dtype(["i", "t", "t", "t", "i", "t"])
        table.fill_from_dict_list(sorted_list)
        table.set_cols_align(["r", "r", "r", "r", "r", "l"])
        table.set_cols_dtype(["i", "t", "t", "t", "i", "t"])
        print(f"\n{table.draw()}")

    else:
        list_issued = []
        for issuer in issuers:
            found = False
            for issued in list_issued:
                if issued.get("uid") is not None and issued["uid"] == issuer["uid"]:
                    issued["blocks"] += 1
                    found = True
                    break
            if not found:
                issued = OrderedDict()
                issued["uid"] = issuer["uid"]
                issued["blocks"] = 1
                list_issued.append(issued)
        for issued in list_issued:
            issued["percent"] = round(issued["blocks"] / number * 100)
        sorted_list = sorted(list_issued, key=itemgetter("blocks"), reverse=True)
        table = tui.Table(style="columns")
        table.fill_from_dict_list(sorted_list)
        table.set_cols_align(["l", "r", "r"])
        table.set_cols_dtype(["t", "i", "i"])
        print(f"from {len(list_issued)} issuers\n{table.draw()}")



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.


import math
import re
import shlex
import time
from typing import List, Optional, Tuple

import click
from duniterpy.api.bma.tx import process
from duniterpy.documents import (
    BlockID,
    InputSource,
    OutputSource,
    SIGParameter,
    Transaction,
    Unlock,
)
from duniterpy.key import SigningKey

from silkaj import auth, network, public_key, tools, tui
from silkaj.blockchain import tools as bc_tools
from silkaj.constants import (
    BMA_SLEEP,
    CENT_MULT_TO_UNIT,
    MINIMAL_ABSOLUTE_TX_AMOUNT,
    MINIMAL_RELATIVE_TX_AMOUNT,
)
from silkaj.money import tools as m_tools
from silkaj.public_key import gen_pubkey_checksum

MAX_COMMENT_LENGTH = 255


# max size for tx doc is 100 lines.
# Formula for accepted field numbers is:
# (2 * IU + 2 * IS + OUT) <= ( MAX_LINES_IN_TX_DOC - FIX_LINES)
# with IU = inputs/unlocks ; IS = Issuers/Signatures ; OUT = Outpouts.
MAX_LINES_IN_TX_DOC = 100
# 2 lines are necessary, and we block 1 more for the comment
FIX_LINES = 3
# assuming there is only 1 issuer and 2 outputs, max inputs is 46
MAX_INPUTS_PER_TX = 46
# assuming there is 1 issuer and 1 input, max outputs is 93.
MAX_OUTPUTS = 93
# for now, silkaj handles txs for one issuer only
NBR_ISSUERS = 1


@click.command("transfer", help="Transfer money")
@click.option(
    "amounts",
    "--amount",
    "-a",
    multiple=True,
    type=click.FloatRange(MINIMAL_ABSOLUTE_TX_AMOUNT),
    help=f"Quantitative amount(s):\n-a <amount>\nMinimum amount is {MINIMAL_ABSOLUTE_TX_AMOUNT}.",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["amountsud", "allsources", "file_path"],
)
@click.option(
    "amountsud",
    "--amountUD",
    "-d",
    multiple=True,
    type=click.FloatRange(MINIMAL_RELATIVE_TX_AMOUNT),
    help=f"Relative amount(s):\n-d <amount_UD>\nMinimum amount is {MINIMAL_RELATIVE_TX_AMOUNT}",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["amounts", "allsources", "file_path"],
)
@click.option(
    "--allSources",
    is_flag=True,
    help="Send all sources to one recipient",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["amounts", "amountsud", "file_path"],
)
@click.option(
    "recipients",
    "--recipient",
    "-r",
    multiple=True,
    help="Pubkey(s)’ recipients + optional checksum:\n-r <pubkey>[:checksum]\n\
Sending to many recipients is possible:\n\
* With one amount, all will receive the amount\n\
* With many amounts (one per recipient)",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["file_path"],
)
@click.option(
    "file_path",
    "--file",
    "-f",
    help="File’s path containing a list of amounts in absolute or \
relative reference and recipients’ pubkeys",
    cls=tools.MutuallyExclusiveOption,
    mutually_exclusive=["recipients", "amounts", "amountsUD", "allsources"],
)
@click.option("--comment", "-c", default="", help="Comment")
@click.option(
    "--outputBackChange",
    help="Pubkey recipient to send the rest of the transaction: <pubkey[:checksum]>",
)
@click.option(
    "--yes", "-y", is_flag=True, help="Assume yes. Do not prompt confirmation"
)
def transfer_money(
    amounts: List[float],
    amountsud: List[float],
    allsources: bool,
    recipients: List[str],
    file_path: str,
    comment: str,
    outputbackchange: str,
    yes: bool,
) -> None:
    if file_path:
        tx_amounts, recipients = parse_file_containing_amounts_recipients(file_path)
    else:
        if not (amounts or amountsud or allsources):
            tools.message_exit("Error: amount, amountUD or allSources is not set.")
        if not recipients:
            tools.message_exit("Error: A recipient should be passed")
        if allsources and len(recipients) > 1:
            tools.message_exit(
                "Error: the --allSources option can only be used with one recipient."
            )
        # compute amounts and amountsud
        if not allsources:
            tx_amounts = transaction_amount(amounts, amountsud, recipients)

    key = auth.auth_method()
    issuer_pubkey = key.pubkey

    pubkey_amount = m_tools.get_amount_from_pubkey(issuer_pubkey)
    if allsources:
        if pubkey_amount[0] <= 0:
            tools.message_exit(
                f"Error: Issuer pubkey {gen_pubkey_checksum(issuer_pubkey)} is empty. \
No transaction sent."
            )

        tx_amounts = [pubkey_amount[0]]

    recipients = list(recipients)
    outputbackchange = check_transaction_values(
        comment,
        recipients,
        outputbackchange,
        pubkey_amount[0] < sum(tx_amounts),
        issuer_pubkey,
    )

    if not yes:
        table = tui.Table()
        table.fill_rows(
            gen_confirmation_table(
                issuer_pubkey,
                pubkey_amount[0],
                tx_amounts,
                recipients,
                outputbackchange,
                comment,
            ),
        )
        confirmation_table = table.draw()

    if yes or click.confirm(
        f"{confirmation_table}\nDo you confirm sending this transaction?"
    ):
        handle_intermediaries_transactions(
            key,
            issuer_pubkey,
            tx_amounts,
            recipients,
            comment,
            outputbackchange,
        )


def parse_file_containing_amounts_recipients(
    file_path: str,
) -> Tuple[List[int], List[str]]:
    """
    Parse file in a specific format
    Comments are ignored
    Format should be:
    ```txt
    [ABSOLUTE/RELATIVE]

    # comment1
    amount1 recipient1’s pubkey
    # comment2
    amount2 recipient2’s pubkey
    ```
    """
    reference = ""
    amounts, recipients = [], []
    with open(file_path, encoding="utf-8") as file:
        for n, raw_line in enumerate(file):
            line = shlex.split(raw_line, True)
            if line:
                if n == 0:
                    reference = line[0]
                else:
                    try:
                        amounts.append(float(line[0]))
                        recipients.append(line[1])
                    except (ValueError, IndexError):
                        tools.message_exit(f"Syntax error at line {n + 1}")

    if not reference or reference not in ("ABSOLUTE", "RELATIVE"):
        tools.message_exit(
            f"{file_path} must contain at first line 'ABSOLUTE' or 'RELATIVE' header"
        )

    if not amounts or not recipients:
        tools.message_exit("No amounts or recipients specified")

    # Compute amount depending on the reference
    if reference == "ABSOLUTE":
        reference_mult = CENT_MULT_TO_UNIT
    else:
        reference_mult = m_tools.get_ud_value()
    tx_amounts = compute_amounts(amounts, reference_mult)

    return tx_amounts, recipients


def transaction_amount(
    amounts: List[float], UDs_amounts: List[float], outputAddresses: List[str]
) -> List[int]:
    """
    Check that the number of passed amounts(UD) and recipients are the same
    Returns a list of amounts.
    """
    # Create amounts list
    if amounts:
        amounts_list = compute_amounts(amounts, CENT_MULT_TO_UNIT)
    elif UDs_amounts:
        UD_value = m_tools.get_ud_value()
        amounts_list = compute_amounts(UDs_amounts, UD_value)
    if len(amounts_list) != len(outputAddresses) and len(amounts_list) != 1:
        tools.message_exit(
            "Error: The number of passed recipients is not the same as the passed amounts."
        )
    # In case one amount is passed with multiple recipients
    # generate list containing multiple time the same amount
    if len(amounts_list) == 1 and len(outputAddresses) > 1:
        amounts_list = [amounts_list[0]] * len(outputAddresses)
    return amounts_list


def compute_amounts(amounts: List[float], multiplicator: float) -> List[int]:
    """
    Computes the amounts(UD) and returns a list.
    Multiplicator should be either CENT_MULT_TO_UNIT or UD_Value.
    If relative amount, check that amount is superior to minimal amount.
    """
    # Create amounts list
    amounts_list = []
    for amount in amounts:
        computed_amount = amount * multiplicator
        # check if relative amounts are high enough
        if (multiplicator != CENT_MULT_TO_UNIT) and (
            computed_amount < (MINIMAL_ABSOLUTE_TX_AMOUNT * CENT_MULT_TO_UNIT)
        ):
            tools.message_exit(f"Error: amount {amount} is too low.")
        amounts_list.append(round(computed_amount))
    return amounts_list


def check_transaction_values(
    comment: str,
    outputAddresses: List[str],
    outputBackChange: str,
    enough_source: bool,
    issuer_pubkey: str,
) -> str:
    """
    Check the comment format
    Check the pubkeys and the checksums of the recipients and the outputbackchange
    In case of a valid checksum, assign and return the pubkey without the checksum
    Check the balance is big enough for the transaction
    """
    checkComment(comment)
    # we check output numbers and leave one line for the backchange.
    if len(outputAddresses) > (MAX_OUTPUTS - 1):
        tools.message_exit(
            f"Error : there should be less than {MAX_OUTPUTS - 1} outputs."
        )
    for i, outputAddress in enumerate(outputAddresses):
        if public_key.check_pubkey_format(outputAddress):
            outputAddresses[i] = public_key.validate_checksum(outputAddress)
    if outputBackChange:
        if public_key.check_pubkey_format(outputBackChange):
            outputBackChange = public_key.validate_checksum(outputBackChange)
    if enough_source:
        pubkey = gen_pubkey_checksum(issuer_pubkey)
        tools.message_exit(
            f"{pubkey} pubkey doesn’t have enough money for this transaction."
        )
    return outputBackChange


def gen_confirmation_table(
    issuer_pubkey: str,
    pubkey_amount: int,
    tx_amounts: List[int],
    outputAddresses: List[str],
    outputBackChange: str,
    comment: str,
) -> List[List]:
    """
    Generate transaction confirmation
    """

    currency_symbol = tools.get_currency_symbol()
    ud_value = m_tools.get_ud_value()
    total_tx_amount = sum(tx_amounts)
    tx = []  # type: List[List[str]]
    # display account situation
    m_tools.display_amount(
        tx,
        "Initial balance",
        pubkey_amount,
        ud_value,
        currency_symbol,
    )
    m_tools.display_amount(
        tx,
        "Total transaction amount",
        total_tx_amount,
        ud_value,
        currency_symbol,
    )
    m_tools.display_amount(
        tx,
        "Balance after transaction",
        (pubkey_amount - total_tx_amount),
        ud_value,
        currency_symbol,
    )
    m_tools.display_pubkey(tx, "From", issuer_pubkey)
    # display outputs and amounts
    for outputAddress, tx_amount in zip(outputAddresses, tx_amounts):
        m_tools.display_pubkey(tx, "To", outputAddress)
        time.sleep(BMA_SLEEP)
        m_tools.display_amount(tx, "Amount", tx_amount, ud_value, currency_symbol)
    # display last informations
    if outputBackChange:
        m_tools.display_pubkey(tx, "Backchange", outputBackChange)
    tx.append(["Comment", comment])
    return tx


def get_list_input_for_transaction(
    pubkey: str, TXamount: int, outputs_number: int
) -> Tuple[List[InputSource], int, bool]:
    listinput = m_tools.get_sources(pubkey)[0]
    maxInputsNumber = max_inputs_number(outputs_number, NBR_ISSUERS)
    # generate final list source
    listinputfinal = []
    totalAmountInput = 0
    intermediatetransaction = False
    for nbr_inputs, _input in enumerate(listinput, start=1):
        listinputfinal.append(_input)
        totalAmountInput += m_tools.amount_in_current_base(_input)
        TXamount -= m_tools.amount_in_current_base(_input)
        # if too much sources, it's an intermediate transaction.
        amount_not_reached_and_max_doc_size_reached = (
            TXamount > 0 and MAX_INPUTS_PER_TX <= nbr_inputs
        )
        amount_reached_too_much_inputs = TXamount <= 0 and maxInputsNumber < nbr_inputs
        if (
            amount_not_reached_and_max_doc_size_reached
            or amount_reached_too_much_inputs
        ):
            intermediatetransaction = True
        # if we reach the MAX_INPUTX_PER_TX limit, we send the interm.tx
        # if we gather the good amount, we send the tx :
        #    - either this is no int.tx, and the tx is sent to the receiver,
        #    - or the int.tx it is sent to the issuer before sent to the receiver.
        if MAX_INPUTS_PER_TX <= nbr_inputs or TXamount <= 0:
            break
    if TXamount > 0 and not intermediatetransaction:
        tools.message_exit("Error: you don't have enough money")
    return listinputfinal, totalAmountInput, intermediatetransaction


def handle_intermediaries_transactions(
    key: SigningKey,
    issuers: str,
    tx_amounts: List[int],
    outputAddresses: List[str],
    Comment: str = "",
    OutputbackChange: Optional[str] = None,
) -> None:
    while True:
        # consider there is always one backchange output, hence +1
        listinput_and_amount = get_list_input_for_transaction(
            issuers, sum(tx_amounts), len(outputAddresses) + 1
        )
        intermediatetransaction = listinput_and_amount[2]

        if intermediatetransaction:
            totalAmountInput = listinput_and_amount[1]
            generate_and_send_transaction(
                key,
                issuers,
                [totalAmountInput],
                listinput_and_amount,
                [issuers],
                "Change operation",
            )
        else:
            generate_and_send_transaction(
                key,
                issuers,
                tx_amounts,
                listinput_and_amount,
                outputAddresses,
                Comment,
                OutputbackChange,
            )
            break


def max_inputs_number(outputs_number: int, issuers_number: int) -> int:
    """
    returns the maximum number of inputs.
    This function does not take care of backchange line.
    formula is IU <= (MAX_LINES_IN_TX_DOC - FIX_LINES - O - 2*IS)/2
    """
    return int(
        (MAX_LINES_IN_TX_DOC - FIX_LINES - (2 * issuers_number) - outputs_number) / 2
    )


def generate_and_send_transaction(
    key: SigningKey,
    issuers: str,
    tx_amounts: List[int],
    listinput_and_amount: Tuple[List[InputSource], int, bool],
    outputAddresses: List[str],
    Comment: str,
    OutputbackChange: Optional[str] = None,
) -> None:
    """
    Display sent transaction
    Generate, sign, and send transaction document
    """
    intermediate_tx = listinput_and_amount[2]
    if intermediate_tx:
        print("Generate Change Transaction")
    else:
        print("Generate Transaction:")
    print("   - From:    " + gen_pubkey_checksum(issuers))
    for tx_amount, outputAddress in zip(tx_amounts, outputAddresses):
        display_sent_tx(outputAddress, tx_amount)
    print("   - Total:   " + str(sum(tx_amounts) / 100))

    transaction = generate_transaction_document(
        issuers,
        tx_amounts,
        listinput_and_amount,
        outputAddresses,
        Comment,
        OutputbackChange,
    )
    transaction.sign(key)
    network.send_document(process, transaction)


def display_sent_tx(outputAddress: str, amount: int) -> None:
    print(
        "   - To:     ",
        gen_pubkey_checksum(outputAddress),
        "\n   - Amount: ",
        amount / 100,
    )


def generate_transaction_document(
    issuers: str,
    tx_amounts: List[int],
    listinput_and_amount: Tuple[List[InputSource], int, bool],
    outputAddresses: List[str],
    Comment: str = "",
    OutputbackChange: Optional[str] = None,
) -> Transaction:

    listinput = listinput_and_amount[0]
    totalAmountInput = listinput_and_amount[1]
    total_tx_amount = sum(tx_amounts)

    head_block = bc_tools.get_head_block()

    if not OutputbackChange:
        OutputbackChange = issuers

    # If it's not a foreign exchange transaction,
    # we remove units after two digits after the decimal point
    if issuers not in outputAddresses:
        total_tx_amount = (
            total_tx_amount // 10 ** head_block["unitbase"]
        ) * 10 ** head_block["unitbase"]

    # Generate output
    ################
    listoutput = []  # type: List[OutputSource]
    for tx_amount, outputAddress in zip(tx_amounts, outputAddresses):
        generate_output(listoutput, head_block["unitbase"], tx_amount, outputAddress)

    # Outputs to himself
    rest = totalAmountInput - total_tx_amount
    generate_output(listoutput, head_block["unitbase"], rest, OutputbackChange)

    # Unlocks
    unlocks = generate_unlocks(listinput)

    # Generate transaction document
    ##############################

    return Transaction(
        block_id=BlockID(head_block["number"], head_block["hash"]),
        locktime=0,
        issuers=[issuers],
        inputs=listinput,
        unlocks=unlocks,
        outputs=listoutput,
        comment=Comment,
        currency=head_block["currency"],
    )


def generate_unlocks(listinput: List[InputSource]) -> List[Unlock]:
    unlocks = []
    for i in range(0, len(listinput)):
        unlocks.append(Unlock(index=i, parameters=[SIGParameter(0)]))
    return unlocks


def generate_output(
    listoutput: List[OutputSource], unitbase: int, rest: int, recipient_address: str
) -> None:
    while rest > 0:
        outputAmount = truncBase(rest, unitbase)
        rest -= outputAmount
        if outputAmount > 0:
            outputAmount = int(outputAmount / math.pow(10, unitbase))
            listoutput.append(
                OutputSource(
                    amount=outputAmount,
                    base=unitbase,
                    condition=f"SIG({recipient_address})",
                )
            )
        unitbase = unitbase - 1


def checkComment(comment: str) -> None:
    if len(comment) > MAX_COMMENT_LENGTH:
        tools.message_exit("Error: Comment is too long")
    regex = re.compile(
        "^[0-9a-zA-Z\\ \\-\\_\\:\\/\\;\\*\\[\\]\\(\\)\\?\
\\!\\^\\+\\=\\@\\&\\~\\#\\{\\}\\|\\\\<\\>\\%\\.]*$"
    )
    if not re.search(regex, comment):
        tools.message_exit("Error: the format of the comment is invalid")


def truncBase(amount: int, base: int) -> int:
    _pow = int(math.pow(10, base))
    if amount < _pow:
        return 0
    return math.trunc(amount / _pow) * _pow



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import re
import sys
from getpass import getpass
from pathlib import Path

from click import Context, command, confirm, option, pass_context
from duniterpy.key import SigningKey
from duniterpy.key.scrypt_params import ScryptParams

from silkaj.constants import FAILURE_EXIT_STATUS, PUBKEY_PATTERN
from silkaj.public_key import gen_pubkey_checksum

SEED_HEX_PATTERN = "^[0-9a-fA-F]{64}$"
PUBSEC_PUBKEY_PATTERN = f"pub: ({PUBKEY_PATTERN})"
PUBSEC_SIGNKEY_PATTERN = "sec: ([1-9A-HJ-NP-Za-km-z]{87,90})"


@pass_context
def auth_method(ctx: Context) -> SigningKey:
    if ctx.obj["AUTH_SEED"]:
        return auth_by_seed()
    if ctx.obj["AUTH_FILE"]:
        return auth_by_auth_file()
    if ctx.obj["AUTH_WIF"]:
        return auth_by_wif()
    return auth_by_scrypt()


@pass_context
def has_auth_method(ctx: Context) -> bool:
    return (
        ctx.obj["AUTH_SCRYPT"]
        or ctx.obj["AUTH_FILE"]
        or ctx.obj["AUTH_SEED"]
        or ctx.obj["AUTH_WIF"]
    )


@command("authentication", help="Generate authentication file")
@option("--file", default="authfile", show_default=True, help="Path file")
def generate_auth_file(file: str) -> None:
    key = auth_method()
    authfile = Path(file)
    pubkey_cksum = gen_pubkey_checksum(key.pubkey)
    if authfile.is_file():
        message = f"Would you like to erase {file} by an authfile corresponding \n\
to following pubkey `{pubkey_cksum}`?"
        confirm(message, abort=True)
    key.save_seedhex_file(file)
    print(
        f"Authentication file 'authfile' generated and stored in current\
 folder for following public key: {pubkey_cksum}",
    )


@pass_context
def auth_by_auth_file(ctx: Context) -> SigningKey:
    """
    Uses an authentication file to generate the key
    Authfile can either be:
    * A seed in hexadecimal encoding
    * PubSec format with public and private key in base58 encoding
    """
    file = ctx.obj["AUTH_FILE_PATH"]
    authfile = Path(file)
    if not authfile.is_file():
        sys.exit(f"Error: `{file}` file does not exist")
    filetxt = authfile.open("r", encoding="utf-8").read()

    # two regural expressions for the PubSec format
    regex_pubkey = re.compile(PUBSEC_PUBKEY_PATTERN, re.MULTILINE)
    regex_signkey = re.compile(PUBSEC_SIGNKEY_PATTERN, re.MULTILINE)

    # Seed hexadecimal format
    if re.search(re.compile(SEED_HEX_PATTERN), filetxt):
        return SigningKey.from_seedhex_file(file)
    # PubSec format
    if re.search(regex_pubkey, filetxt) and re.search(regex_signkey, filetxt):
        return SigningKey.from_pubsec_file(file)
    sys.exit("Error: the format of the file is invalid")


def auth_by_seed() -> SigningKey:
    seedhex = getpass("Please enter your seed on hex format: ")
    try:
        return SigningKey.from_seedhex(seedhex)
    # To be fixed upstream in DuniterPy
    # pylint: disable=broad-except
    except Exception as error:
        print(error)
        sys.exit(FAILURE_EXIT_STATUS)


@pass_context
def auth_by_scrypt(ctx: Context) -> SigningKey:
    salt = getpass("Please enter your Scrypt Salt (Secret identifier): ")
    password = getpass("Please enter your Scrypt password (masked): ")

    if ctx.obj["AUTH_SCRYPT_PARAMS"]:
        n, r, p = ctx.obj["AUTH_SCRYPT_PARAMS"].split(",")

        if n.isnumeric() and r.isnumeric() and p.isnumeric():
            # pylint: disable=too-many-boolean-expressions
            n, r, p = int(n), int(r), int(p)
            if n <= 0 or n > 65536 or r <= 0 or r > 512 or p <= 0 or p > 32:
                sys.exit("Error: the values of Scrypt parameters are not good")
            scrypt_params = ScryptParams(n, r, p)
        else:
            sys.exit("one of n, r or p is not a number")
    else:
        scrypt_params = None

    try:
        return SigningKey.from_credentials(salt, password, scrypt_params)
    except ValueError as error:
        print(error)
        sys.exit(FAILURE_EXIT_STATUS)


def auth_by_wif() -> SigningKey:
    wif_hex = getpass("Enter your WIF or Encrypted WIF address (masked): ")
    password = getpass(
        "(Leave empty in case WIF format) Enter the Encrypted WIF password (masked): "
    )
    try:
        return SigningKey.from_wif_or_ewif_hex(wif_hex, password)
    # To be fixed upstream in DuniterPy
    # pylint: disable=broad-except
    except Exception as error:
        print(error)
        sys.exit(FAILURE_EXIT_STATUS)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import time
import urllib
from typing import Dict, List, Optional, Tuple
from urllib.error import HTTPError

import click
from duniterpy.api.bma import wot

from silkaj.constants import BMA_SLEEP
from silkaj.network import client_instance, exit_on_http_error
from silkaj.public_key import gen_pubkey_checksum
from silkaj.tui import Table


def identity_of(pubkey_uid: str) -> Dict:
    """
    Only works for members
    Not able to get corresponding uid from a non-member identity
    Able to know if an identity is member or not
    """
    client = client_instance()
    return client(wot.identity_of, pubkey_uid)


def is_member(pubkey_uid: str) -> Optional[Dict]:
    """
    Check identity is member
    If member, return corresponding identity, else: False
    """
    try:
        return identity_of(pubkey_uid)
    except HTTPError:
        return None


def wot_lookup(identifier: str) -> List:
    """
    :identifier: identity or pubkey in part or whole
    Return received and sent certifications lists of matching identities
    if one identity found
    """
    client = client_instance()
    return (client(wot.lookup, identifier))["results"]


def identities_from_pubkeys(pubkeys: List[str], uids: bool) -> List:
    """
    Make list of pubkeys unique, and remove empty strings
    Request identities
    """
    if not uids:
        return []

    uniq_pubkeys = list(filter(None, set(pubkeys)))
    identities = []
    for pubkey in uniq_pubkeys:
        time.sleep(BMA_SLEEP)
        try:
            identities.append(identity_of(pubkey))
        except HTTPError:
            pass
    return identities


def choose_identity(pubkey_uid: str) -> Tuple[Dict, str, List]:
    """
    Get lookup from a pubkey or an uid
    Loop over the double lists: pubkeys, then uids
    If there is one uid, returns it
    If there is multiple uids, prompt a selector
    """

    try:
        lookups = wot_lookup(pubkey_uid)
    except urllib.error.HTTPError as e:
        exit_on_http_error(e, 404, f"No identity found for {pubkey_uid}")

    # Generate table containing the choices
    identities_choices = {
        "id": [],
        "uid": [],
        "pubkey": [],
        "timestamp": [],
    }  # type: Dict
    for pubkey_index, lookup in enumerate(lookups):
        for uid_index, identity in enumerate(lookup["uids"]):
            identities_choices["id"].append(str(pubkey_index) + str(uid_index))
            identities_choices["pubkey"].append(gen_pubkey_checksum(lookup["pubkey"]))
            identities_choices["uid"].append(identity["uid"])
            identities_choices["timestamp"].append(
                identity["meta"]["timestamp"][:20] + "…"
            )

    identities = len(identities_choices["uid"])
    if identities == 1:
        pubkey_index = 0
        uid_index = 0
    elif identities > 1:
        table = Table().set_cols_dtype(["t", "t", "t", "t"])
        table.fill_from_dict(identities_choices)
        click.echo(table.draw())
        # Loop till the passed value is in identities_choices
        message = "Which identity would you like to select (id)?"
        selected_id = None
        while selected_id not in identities_choices["id"]:
            selected_id = click.prompt(message)

        pubkey_index = int(str(selected_id)[:-1])
        uid_index = int(str(selected_id)[-1:])

    return (
        lookups[pubkey_index]["uids"][uid_index],
        lookups[pubkey_index]["pubkey"],
        lookups[pubkey_index]["signed"],
    )



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import logging
import sys

import click
import pendulum
from duniterpy.api import bma
from duniterpy.documents import BlockID, Membership, get_block_id
from duniterpy.key import SigningKey

from silkaj import auth, tui
from silkaj.blockchain import tools as bc_tools
from silkaj.constants import DATE, SUCCESS_EXIT_STATUS
from silkaj.g1_monetary_license import license_approval
from silkaj.network import client_instance, send_document
from silkaj.public_key import gen_pubkey_checksum
from silkaj.wot import tools as w_tools


@click.command(
    "membership",
    help="Send and sign membership document: \n\
for first emission and for renewal",
)
@click.pass_context
def send_membership(ctx: click.Context) -> None:
    dry_run = ctx.obj["DRY_RUN"]

    # Authentication
    key = auth.auth_method()

    # Get the identity information
    head_block = bc_tools.get_head_block()
    membership_block_id = BlockID(head_block["number"], head_block["hash"])
    identity = (w_tools.choose_identity(key.pubkey))[0]
    identity_uid = identity["uid"]
    identity_block_id = get_block_id(identity["meta"]["timestamp"])

    # Display license and ask for confirmation
    currency = head_block["currency"]
    if not dry_run:
        license_approval(currency)

    # Confirmation
    display_confirmation_table(identity_uid, key.pubkey, identity_block_id)
    if not dry_run and not ctx.obj["DISPLAY_DOCUMENT"]:
        tui.send_doc_confirmation("membership document for this identity")

    # Create and sign membership document
    membership = generate_membership_document(
        key.pubkey,
        membership_block_id,
        identity_uid,
        identity_block_id,
        currency,
        key,
    )

    logging.debug(membership.signed_raw())

    if dry_run:
        click.echo(membership.signed_raw())
        sys.exit(SUCCESS_EXIT_STATUS)

    if ctx.obj["DISPLAY_DOCUMENT"]:
        click.echo(membership.signed_raw())
        tui.send_doc_confirmation("membership document for this identity")

    # Send the membership signed raw document to the node
    send_document(bma.blockchain.membership, membership)


def display_confirmation_table(
    identity_uid: str, pubkey: str, identity_block_id: BlockID
) -> None:
    """
    Check whether there is pending memberships already in the mempool
    Display their expiration date

    Actually, sending a membership document works even if the time
    between two renewals is not awaited as for the certification
    """

    client = client_instance()

    identities_requirements = client(bma.wot.requirements, pubkey)
    for identity_requirements in identities_requirements["identities"]:
        if identity_requirements["uid"] == identity_uid:
            membership_expires = identity_requirements["membershipExpiresIn"]
            pending_expires = identity_requirements["membershipPendingExpiresIn"]
            pending_memberships = identity_requirements["pendingMemberships"]
            break

    table = []
    if membership_expires:
        expires = pendulum.now().add(seconds=membership_expires).diff_for_humans()
        table.append(["Expiration date of current membership", expires])

    if pending_memberships:
        table.append(
            [
                "Number of pending membership(s) in the mempool",
                len(pending_memberships),
            ]
        )

        table.append(
            [
                "Pending membership documents will expire",
                pendulum.now().add(seconds=pending_expires).diff_for_humans(),
            ]
        )

    table.append(["User Identifier (UID)", identity_uid])
    table.append(["Public Key", gen_pubkey_checksum(pubkey)])

    table.append(["Block Identity", str(identity_block_id)[:45] + "…"])

    block = client(bma.blockchain.block, identity_block_id.number)
    table.append(
        [
            "Identity published",
            pendulum.from_timestamp(block["time"], tz="local").format(DATE),
        ]
    )

    params = bc_tools.get_blockchain_parameters()
    table.append(
        [
            "Expiration date of new membership",
            pendulum.now().add(seconds=params["msValidity"]).diff_for_humans(),
        ]
    )

    table.append(
        [
            "Expiration date of new membership from the mempool",
            pendulum.now().add(seconds=params["msPeriod"]).diff_for_humans(),
        ]
    )

    display_table = tui.Table()
    display_table.fill_rows(table)
    click.echo(display_table.draw())


def generate_membership_document(
    pubkey: str,
    membership_block_id: BlockID,
    identity_uid: str,
    identity_block_id: BlockID,
    currency: str,
    key: SigningKey = None,
) -> Membership:
    return Membership(
        issuer=pubkey,
        membership_block_id=membership_block_id,
        uid=identity_uid,
        identity_block_id=identity_block_id,
        currency=currency,
        signing_key=key,
    )



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

from operator import eq, itemgetter, ne, neg
from typing import Any, List, Optional, Tuple
from urllib.error import HTTPError

from click import argument, command, echo_via_pager, option
from duniterpy.api.bma.tx import history
from duniterpy.api.client import Client
from duniterpy.documents.transaction import OutputSource, Transaction
from duniterpy.grammars.output import Condition
from pendulum import from_timestamp, now

from silkaj.constants import ALL, ALL_DIGITAL
from silkaj.money.tools import (
    amount_in_current_base,
    get_amount_from_pubkey,
    get_ud_value,
)
from silkaj.network import client_instance
from silkaj.public_key import (
    check_pubkey_format,
    gen_pubkey_checksum,
    validate_checksum,
)
from silkaj.tools import get_currency_symbol
from silkaj.tui import Table
from silkaj.wot import tools as wt


@command("history", help="Display transaction history")
@argument("pubkey")
@option("--uids", "-u", is_flag=True, help="Display uids")
@option("--full-pubkey", "-f", is_flag=True, help="Display full-length pubkeys")
def transaction_history(pubkey: str, uids: bool, full_pubkey: bool) -> None:
    if check_pubkey_format(pubkey):
        pubkey = validate_checksum(pubkey)

    client = client_instance()
    ud_value = get_ud_value()
    currency_symbol = get_currency_symbol()

    header = generate_header(pubkey, currency_symbol, ud_value)
    received_txs, sent_txs = [], []  # type: List[Transaction], List[Transaction]
    get_transactions_history(client, pubkey, received_txs, sent_txs)
    remove_duplicate_txs(received_txs, sent_txs)

    txs_list = generate_txs_list(
        received_txs, sent_txs, pubkey, ud_value, currency_symbol, uids, full_pubkey
    )
    table_headers = [
        "Date",
        "Issuers/Recipients",
        f"Amounts {currency_symbol}",
        f"Amounts UD{currency_symbol}",
        "Comment",
    ]
    table = Table()
    table.fill_rows(txs_list, table_headers)
    echo_via_pager(header + table.draw())


def generate_header(pubkey: str, currency_symbol: str, ud_value: int) -> str:
    try:
        idty = wt.identity_of(pubkey)
    except HTTPError:
        idty = dict([("uid", "")])
    balance = get_amount_from_pubkey(pubkey)
    balance_ud = round(balance[1] / ud_value, 2)
    date = now().format(ALL)
    return f'Transactions history from: {idty["uid"]} {gen_pubkey_checksum(pubkey)}\n\
Current balance: {balance[1] / 100} {currency_symbol}, {balance_ud} UD {currency_symbol} on {date}\n'


def get_transactions_history(
    client: Client, pubkey: str, received_txs: List, sent_txs: List
) -> None:
    """
    Get transaction history
    Store txs in Transaction object
    """
    tx_history = client(history, pubkey)
    currency = tx_history["currency"]

    for received in tx_history["history"]["received"]:
        received_txs.append(Transaction.from_bma_history(received, currency))
    for sent in tx_history["history"]["sent"]:
        sent_txs.append(Transaction.from_bma_history(sent, currency))


def remove_duplicate_txs(received_txs: List, sent_txs: List) -> None:
    """
    Remove duplicate transactions from history
    Remove received tx which contains output back return
    that we don’t want to displayed
    A copy of received_txs is necessary to remove elements
    """
    for received_tx in list(received_txs):
        if received_tx in sent_txs:
            received_txs.remove(received_tx)


def generate_txs_list(
    received_txs: List[Transaction],
    sent_txs: List[Transaction],
    pubkey: str,
    ud_value: int,
    currency_symbol: str,
    uids: bool,
    full_pubkey: bool,
) -> List:
    """
    Generate information in a list of lists for texttable
    Merge received and sent txs
    Sort txs temporarily
    """

    received_txs_list, sent_txs_list = (
        [],
        [],
    )  # type: List[Transaction], List[Transaction]
    parse_received_tx(
        received_txs_list, received_txs, pubkey, ud_value, uids, full_pubkey
    )
    parse_sent_tx(sent_txs_list, sent_txs, pubkey, ud_value, uids, full_pubkey)
    txs_list = received_txs_list + sent_txs_list

    txs_list.sort(key=itemgetter(0), reverse=True)
    return txs_list


def parse_received_tx(
    received_txs_table: List[Transaction],
    received_txs: List[Transaction],
    pubkey: str,
    ud_value: int,
    uids: bool,
    full_pubkey: bool,
) -> None:
    """
    Extract issuers’ pubkeys
    Get identities from pubkeys
    Convert time into human format
    Assign identities
    Get amounts and assign amounts and amounts_ud
    Append comment
    """
    issuers = []
    for received_tx in received_txs:
        for issuer in received_tx.issuers:
            issuers.append(issuer)
    identities = wt.identities_from_pubkeys(issuers, uids)
    for received_tx in received_txs:
        tx_list = []
        tx_list.append(from_timestamp(received_tx.time, tz="local").format(ALL_DIGITAL))
        tx_list.append("")
        for i, issuer in enumerate(received_tx.issuers):
            tx_list[1] += prefix(None, None, i) + assign_idty_from_pubkey(
                issuer, identities, full_pubkey
            )
        amounts = tx_amount(received_tx, pubkey, received_func)[0]
        tx_list.append(amounts / 100)
        tx_list.append(amounts / ud_value)
        tx_list.append(received_tx.comment)
        received_txs_table.append(tx_list)


def parse_sent_tx(
    sent_txs_table: List[Transaction],
    sent_txs: List[Transaction],
    pubkey: str,
    ud_value: int,
    uids: bool,
    full_pubkey: bool,
) -> None:
    # pylint: disable=too-many-locals
    """
    Extract recipients’ pubkeys from outputs
    Get identities from pubkeys
    Convert time into human format
    Store "Total" and total amounts according to the number of outputs
    If not output back return:
    Assign amounts, amounts_ud, identities, and comment
    """
    pubkeys = []
    for sent_tx in sent_txs:
        outputs = tx_amount(sent_tx, pubkey, sent_func)[1]
        for output in outputs:
            if output_available(output.condition, ne, pubkey):
                pubkeys.append(output.condition.left.pubkey)

    identities = wt.identities_from_pubkeys(pubkeys, uids)
    for sent_tx in sent_txs:
        tx_list = []
        tx_list.append(from_timestamp(sent_tx.time, tz="local").format(ALL_DIGITAL))

        total_amount, outputs = tx_amount(sent_tx, pubkey, sent_func)
        if len(outputs) > 1:
            tx_list.append("Total")
            amounts = str(total_amount / 100)
            amounts_ud = str(round(total_amount / ud_value, 2))
        else:
            tx_list.append("")
            amounts = ""
            amounts_ud = ""

        for i, output in enumerate(outputs):
            if output_available(output.condition, ne, pubkey):
                amounts += prefix(None, outputs, i) + str(
                    neg(amount_in_current_base(output)) / 100
                )
                amounts_ud += prefix(None, outputs, i) + str(
                    round(neg(amount_in_current_base(output)) / ud_value, 2)
                )
                tx_list[1] += prefix(tx_list[1], outputs, 0) + assign_idty_from_pubkey(
                    output.condition.left.pubkey, identities, full_pubkey
                )
        tx_list.append(amounts)
        tx_list.append(amounts_ud)
        tx_list.append(sent_tx.comment)
        sent_txs_table.append(tx_list)


def tx_amount(
    tx: List[Transaction], pubkey: str, function: Any
) -> Tuple[int, List[OutputSource]]:
    """
    Determine transaction amount from output sources
    """
    amount = 0
    outputs = []
    for output in tx.outputs:  # type: ignore
        if output_available(output.condition, ne, pubkey):
            outputs.append(output)
        amount += function(output, pubkey)
    return amount, outputs


def received_func(output: OutputSource, pubkey: str) -> int:
    if output_available(output.condition, eq, pubkey):
        return amount_in_current_base(output)
    return 0


def sent_func(output: OutputSource, pubkey: str) -> int:
    if output_available(output.condition, ne, pubkey):
        return neg(amount_in_current_base(output))
    return 0


def output_available(condition: Condition, comparison: Any, value: str) -> bool:
    """
    Check if output source is available
    Currently only handle simple SIG condition
    XHX, CLTV, CSV should be handled when present in the blockchain
    """
    if hasattr(condition.left, "pubkey"):
        return comparison(condition.left.pubkey, value)
    return False


def assign_idty_from_pubkey(pubkey: str, identities: List, full_pubkey: bool) -> str:
    idty = gen_pubkey_checksum(pubkey, short=not full_pubkey)
    for identity in identities:
        if pubkey == identity["pubkey"]:
            pubkey_mod = gen_pubkey_checksum(pubkey, short=not full_pubkey)
            idty = f'{identity["uid"]} - {pubkey_mod}'
    return idty


def prefix(
    tx_addresses: Optional[str], outputs: Optional[List[OutputSource]], occurence: int
) -> str:
    """
    Pretty print with texttable
    Break line when several values in a cell

    Received tx case, 'outputs' is not defined, then add a breakline
    between the pubkeys except for the first occurence for multi-sig support

    Sent tx case, handle "Total" line in case of multi-output txs
    In case of multiple outputs, there is a "Total" on the top,
    where there must be a breakline
    """

    if not outputs:
        return "\n" if occurence > 0 else ""

    if tx_addresses == "Total":
        return "\n"
    return "\n" if len(outputs) > 1 else ""



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import click

from silkaj.constants import SILKAJ_VERSION


@click.command("about", help="Display program information")
def about() -> None:
    print(
        "\
\n             @@@@@@@@@@@@@\
\n         @@@     @         @@@\
\n      @@@   @@       @@@@@@   @@.            Silkaj",
        SILKAJ_VERSION,
        "\
\n     @@  @@@       @@@@@@@@@@@  @@,\
\n   @@  @@@       &@@@@@@@@@@@@@  @@@         Powerfull and lightweight command line client\
\n  @@  @@@       @@@@@@@@@#   @@@@ @@(\
\n  @@ @@@@      @@@@@@@@@      @@@  @@        Built in Python for Duniter’s currencies: \
Ğ1 and Ğ1-Test\
\n @@  @@@      @@@@@@@@ @       @@@  @@\
\n @@  @@@      @@@@@@ @@@@       @@  @@       Authors: see AUTHORS.md file\
\n @@  @@@@      @@@ @@@@@@@      @@  @@\
\n  @@ @@@@*       @@@@@@@@@      @# @@        Website: https://silkaj.duniter.org\
\n  @@  @@@@@    @@@@@@@@@@       @ ,@@\
\n   @@  @@@@@ @@@@@@@@@@        @ ,@@         Repository: \
https://git.duniter.org/clients/python/silkaj\
\n    @@@  @@@@@@@@@@@@        @  @@*\
\n      @@@  @@@@@@@@        @  @@@            License: GNU AGPLv3\
\n        @@@@   @@          @@@,\
\n            @@@@@@@@@@@@@@@\n",
    )



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

from collections import OrderedDict
from operator import itemgetter
from os import system
from typing import Dict, Tuple

import click
import jsonschema
from duniterpy.api import bma
from duniterpy.api.client import WSConnection
from pendulum import from_timestamp
from websocket._exceptions import WebSocketConnectionClosedException

from silkaj import network, tui
from silkaj.constants import ALL


@click.command(
    "difficulty",
    help="Display the current Proof of Work difficulty level to generate the next block",
)
def difficulties() -> None:
    client = network.client_instance()
    try:
        ws = client(bma.ws.block)
        while True:
            current = ws.receive_json()
            jsonschema.validate(current, bma.ws.WS_BLOCK_SCHEMA)
            diffi = client(bma.blockchain.difficulties)
            display_diffi(current, diffi)
    except (jsonschema.ValidationError, WebSocketConnectionClosedException) as e:
        print(f"{str(e.__class__.__name__)}: {str(e)}")


def display_diffi(current: WSConnection, diffi: Dict) -> None:
    levels = [OrderedDict((i, d[i]) for i in ("uid", "level")) for d in diffi["levels"]]
    diffi["levels"] = levels
    issuers = 0
    sorted_diffi = sorted(diffi["levels"], key=itemgetter("level"), reverse=True)
    for d in diffi["levels"]:
        if d["level"] / 2 < current["powMin"]:
            issuers += 1
        d["match"] = match_pattern(d["level"])[0][:20]
        d["Π diffi"] = compute_power(match_pattern(d["level"])[1])
        d["Σ diffi"] = d.pop("level")
    system("cls||clear")
    block_gen = from_timestamp(current["time"], tz="local").format(ALL)
    match = match_pattern(int(current["powMin"]))[0]

    table = tui.Table(style="columns").set_cols_dtype(["t", "t", "t", "i"])
    table.fill_from_dict_list(sorted_diffi)

    content = f'Current block: n°{current["number"]}, generated on {block_gen}\n\
Generation of next block n°{diffi["block"]} \
possible by at least {issuers}/{len(diffi["levels"])} members\n\
Common Proof-of-Work difficulty level: {current["powMin"]}, hash starting with `{match}`\n\
{table.draw()}'
    print(content)


def match_pattern(_pow: int, match: str = "", p: int = 1) -> Tuple[str, int]:
    while _pow > 0:
        if _pow >= 16:
            match += "0"
            _pow -= 16
            p *= 16
        else:
            match += f"[0-{hex(15 - _pow)[2:].upper()}]"
            p *= _pow
            _pow = 0
    return f"{match}*", p


def compute_power(nbr: float, power: int = 0) -> str:
    while nbr >= 10:
        nbr /= 10
        power += 1
    return f"{nbr:.1f} × 10^{power}"



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import logging
from typing import List
from urllib.error import HTTPError

from click import INT, argument, command, progressbar
from duniterpy.api import bma
from duniterpy.api.client import Client
from duniterpy.documents import Block

from silkaj.constants import BMA_MAX_BLOCKS_CHUNK_SIZE
from silkaj.network import client_instance
from silkaj.tools import message_exit


@command(
    "verify",
    help="Verify blocks’ signatures. \
If only FROM_BLOCK is specified, it verifies from this block to the last block. \
If nothing specified, the whole blockchain gets verified.",
)
@argument("from_block", default=0, type=INT)
@argument("to_block", default=0, type=INT)
def verify_blocks_signatures(from_block: int, to_block: int) -> None:
    client = client_instance()
    to_block = check_passed_blocks_range(client, from_block, to_block)
    invalid_blocks_signatures = []  # type: List[int]
    chunks_from = range(from_block, to_block + 1, BMA_MAX_BLOCKS_CHUNK_SIZE)
    with progressbar(chunks_from, label="Processing blocks verification") as bar:
        for chunk_from in bar:
            chunk_size = get_chunk_size(from_block, to_block, chunks_from, chunk_from)
            logging.info(
                "Processing chunk from block %d to %d",
                chunk_from,
                chunk_from + chunk_size,
            )
            chunk = get_chunk(client, chunk_size, chunk_from)

            for block in chunk:
                block = Block.from_signed_raw(f'{block["raw"]}{block["signature"]}\n')
                verify_block_signature(invalid_blocks_signatures, block)

    display_result(from_block, to_block, invalid_blocks_signatures)


def check_passed_blocks_range(client: Client, from_block: int, to_block: int) -> int:
    head_number = (client(bma.blockchain.current))["number"]
    if to_block == 0:
        to_block = head_number
    if to_block > head_number:
        message_exit(
            f"Passed TO_BLOCK argument is bigger than the head block: {str(head_number)}"
        )
    if from_block > to_block:
        message_exit("TO_BLOCK should be bigger or equal to FROM_BLOCK")
    return to_block


def get_chunk_size(
    from_block: int, to_block: int, chunks_from: range, chunk_from: int
) -> int:
    """If not last chunk, take the maximum size
    Otherwise, calculate the size for the last chunk"""
    if chunk_from != chunks_from[-1]:
        return BMA_MAX_BLOCKS_CHUNK_SIZE
    return (to_block + 1 - from_block) % BMA_MAX_BLOCKS_CHUNK_SIZE


def get_chunk(client: Client, chunk_size: int, chunk_from: int) -> List:
    try:
        chunk = client(bma.blockchain.blocks, chunk_size, chunk_from)
    except HTTPError as e:
        logging.error(e)
        message_exit("Error: Network error to get chunck")
    return chunk


def verify_block_signature(invalid_blocks_signatures: List[int], block: Block) -> None:
    if not block.check_signature(block.issuer):
        invalid_blocks_signatures.append(block.number)


def display_result(
    from_block: int, to_block: int, invalid_blocks_signatures: List[int]
) -> None:
    result = f"Within {from_block}-{to_block} range, "
    if invalid_blocks_signatures:
        result += "blocks with a wrong signature: "
        result += " ".join(str(n) for n in invalid_blocks_signatures)
    else:
        result += "no blocks with a wrong signature."
    print(result)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import hashlib
import re
from typing import Any, Optional, Union

import base58

from silkaj.constants import PUBKEY_PATTERN, SHORT_PUBKEY_SIZE
from silkaj.tools import message_exit

PUBKEY_DELIMITED_PATTERN = f"^{PUBKEY_PATTERN}$"
CHECKSUM_SIZE = 3
CHECKSUM_PATTERN = f"[1-9A-HJ-NP-Za-km-z]{{{CHECKSUM_SIZE}}}"
PUBKEY_CHECKSUM_PATTERN = f"^{PUBKEY_PATTERN}:{CHECKSUM_PATTERN}$"


def is_pubkey_and_check(pubkey: str) -> Union[str, bool]:
    """
    Checks if the given argument contains a pubkey.
    If so, verifies the checksum if needed and returns the pubkey.
    Exits if the checksum is wrong.
    Else, return False
    """
    if re.search(re.compile(PUBKEY_PATTERN), pubkey):
        if check_pubkey_format(pubkey, True):
            return validate_checksum(pubkey)
        return pubkey
    return False


def check_pubkey_format(pubkey: str, display_error: bool = True) -> Optional[bool]:
    """
    Checks if a pubkey has a checksum.
    Exits if the pubkey is invalid.
    """
    if re.search(re.compile(PUBKEY_DELIMITED_PATTERN), pubkey):
        return False
    if re.search(re.compile(PUBKEY_CHECKSUM_PATTERN), pubkey):
        return True
    if display_error:
        message_exit(f"Error: bad format for following public key: {pubkey}")
    return None


def validate_checksum(pubkey_checksum: str) -> Any:
    """
    Check pubkey checksum after the pubkey, delimited by ":".
    If check pass: return pubkey
    Else: exit.
    """
    pubkey, checksum = pubkey_checksum.split(":")
    if checksum == gen_checksum(pubkey):
        return pubkey
    message_exit(
        f"Error: public key '{pubkey}' does not match checksum '{checksum}'.\n\
Please verify the public key."
    )
    return None


def gen_checksum(pubkey: str) -> str:
    """
    Returns the checksum of the input pubkey (encoded in b58)
    """
    pubkey_byte = base58.b58decode(pubkey)
    _hash = hashlib.sha256(hashlib.sha256(pubkey_byte).digest()).digest()
    return str(base58.b58encode(_hash)[:3].decode("utf-8"))


def gen_pubkey_checksum(
    pubkey: str,
    short: Optional[bool] = False,
    length: Optional[int] = SHORT_PUBKEY_SIZE,
) -> str:
    """
    Returns "<pubkey>:<checksum>" in full form.
    returns `length` first chars of pubkey and checksum in short form.
    `length` defaults to SHORT_PUBKEY_SIZE.
    """
    short_pubkey = f"{pubkey[:length]}…" if short else pubkey
    return f"{short_pubkey}:{gen_checksum(pubkey)}"



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import re
import sys

import click

from silkaj.auth import auth_method, has_auth_method
from silkaj.public_key import (
    PUBKEY_CHECKSUM_PATTERN,
    PUBKEY_DELIMITED_PATTERN,
    gen_checksum,
    gen_pubkey_checksum,
)

MESSAGE = "You should specify a pubkey or an authentication method"


@click.command(
    "checksum",
    help="Generate checksum out of a passed pubkey or an authentication method.\
 Can also check if the checksum is valid",
)
@click.argument("pubkey_checksum", nargs=-1)
def checksum_command(pubkey_checksum: str) -> None:
    if has_auth_method():
        key = auth_method()
        click.echo(gen_pubkey_checksum(key.pubkey))
    else:
        if not pubkey_checksum:
            sys.exit(MESSAGE)
        elif re.search(re.compile(PUBKEY_DELIMITED_PATTERN), pubkey_checksum[0]):
            click.echo(gen_pubkey_checksum(pubkey_checksum[0]))
        elif re.search(re.compile(PUBKEY_CHECKSUM_PATTERN), pubkey_checksum[0]):
            pubkey, checksum = pubkey_checksum[0].split(":")
            if checksum == gen_checksum(pubkey):
                click.echo("The checksum is valid")
            else:
                click.echo("The checksum is invalid")
        else:
            sys.exit("Error: Wrong public key format")



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

from collections import OrderedDict
from typing import Dict, List, Tuple

import click
from duniterpy.api.bma import blockchain, wot
from pendulum import from_timestamp, now

from silkaj.blockchain.tools import get_blockchain_parameters
from silkaj.constants import DATE
from silkaj.network import client_instance
from silkaj.public_key import gen_pubkey_checksum, is_pubkey_and_check
from silkaj.tui import Table
from silkaj.wot import tools as wt


def get_sent_certifications(
    signed: List, time_first_block: int, params: Dict
) -> Tuple[List[str], List[str]]:
    sent = []
    expire = []
    if signed:
        for cert in signed:
            sent.append(cert["uid"])
            expire.append(
                expiration_date_from_block_id(
                    cert["cert_time"]["block"], time_first_block, params
                )
            )
    return sent, expire


@click.command(
    "status",
    help="Check received and sent certifications and \
consult the membership status of any given identity",
)
@click.argument("uid_pubkey")
def status(uid_pubkey: str) -> None:
    """
    get searched id
    get id of received and sent certifications
    display in a table the result with the numbers
    """
    client = client_instance()
    first_block = client(blockchain.block, 1)
    time_first_block = first_block["time"]

    checked_pubkey = is_pubkey_and_check(uid_pubkey)
    if checked_pubkey:
        uid_pubkey = str(checked_pubkey)

    identity, pubkey, signed = wt.choose_identity(uid_pubkey)
    certifications = OrderedDict()  # type: OrderedDict
    params = get_blockchain_parameters()

    req = None
    requirements = client(wot.requirements, pubkey)
    for req in requirements["identities"]:
        if req["pubkey"] == pubkey:
            break

    certifications["received_expire"] = []
    certifications["received"] = []
    certifications["sent"] = []
    certifications["sent_expire"] = []
    for cert in identity["others"]:
        certifications["received_expire"].append(
            expiration_date_from_block_id(
                cert["meta"]["block_number"], time_first_block, params
            )
        )
        certifications["received"].append(
            cert_written_in_the_blockchain(req["certifications"], cert)
        )
        (
            certifications["sent"],
            certifications["sent_expire"],
        ) = get_sent_certifications(signed, time_first_block, params)
    nbr_sent_certs = len(certifications["sent"]) if "sent" in certifications else 0

    table = Table(style="columns").set_cols_align(["r", "r", "r", "r"])
    table.fill_from_dict(certifications)

    print(
        f'{identity["uid"]} ({gen_pubkey_checksum(pubkey, True)}) \
from block #{identity["meta"]["timestamp"][:15]}…\n\
received {len(certifications["received"])} and sent \
{nbr_sent_certs}/{params["sigStock"]} certifications:\n\
{table.draw()}\n\
✔: Certification available to be written or already written into the blockchain\n'
    )
    membership_status(certifications, pubkey, req)


def cert_written_in_the_blockchain(written_certs: Dict, certifieur: Dict):
    for cert in written_certs:
        if cert["from"] == certifieur["pubkey"]:
            return certifieur["uids"][0] + " ✔"
    return certifieur["uids"][0]


def membership_status(certifications: OrderedDict, pubkey: str, req: Dict) -> None:
    params = get_blockchain_parameters()
    if len(certifications["received"]) >= params["sigQty"]:
        date = certifications["received_expire"][
            len(certifications["received"]) - params["sigQty"]
        ]
        print(f"Membership expiration due to certification expirations: {date}")
    member_lookup = wt.is_member(pubkey)
    is_member = bool(member_lookup)
    print("member:", is_member)
    if req["revoked"]:
        revoke_date = from_timestamp(req["revoked_on"], tz="local").format(DATE)
        print(f"revoked: {req['revoked']}\nrevoked on: {revoke_date}")
    if not is_member and req["wasMember"]:
        print("expired:", req["expired"], "\nwasMember:", req["wasMember"])
    elif is_member:
        expiration_date = now().add(seconds=req["membershipExpiresIn"]).format(DATE)
        print(f"Membership document expiration: {expiration_date}")
        print("Sentry:", req["isSentry"])
    print("outdistanced:", req["outdistanced"])


def expiration_date_from_block_id(
    block_id: str, time_first_block: int, params: Dict
) -> str:
    expir_timestamp = (
        date_approximation(block_id, time_first_block, params["avgGenTime"])
        + params["sigValidity"]
    )
    return from_timestamp(expir_timestamp, tz="local").format(DATE)


def date_approximation(block_id, time_first_block, avgentime):
    return time_first_block + block_id * avgentime
